/* 
 * Copyright (c) 2009, 2014, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of the
 * License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */

#include "stdafx.h"

#include "mforms/mforms.h"

#define SIDE_PADDING 12
#define TAB_SIDE_PADDING 10
#define TAB_TEXT_SPACING 8

#ifdef __APPLE__
  #define TAB_FONT "Lucida Grande"
#elif _WIN32
  #define TAB_FONT "Tahoma"
#else
  #define TAB_FONT "Helvetica"
#endif
#define TITLE_FONT_SIZE 12
#define SUB_TITLE_FONT_SIZE 9
#define TITLE_TEXT_SPACING 4

#define ICON_WIDTH 32
#define ICON_HEIGHT 32

#define INITIAL_TAB_HEIGHT 58


#ifdef notdef
#define VERTICAL_STYLE_WIDTH 50
#define VERTICAL_STYLE_HEIGHT 60
#else
#define VERTICAL_STYLE_WIDTH 64
#define VERTICAL_STYLE_HEIGHT 70
#endif

#ifndef _WIN32
using std::max;
#endif

using namespace base;
using namespace mforms;


//--------------------------------------------------------------------------------------------------

TabSwitcher::TabSwitcher(TabSwitcherType type)
: _type(type), _tabView(0), _selected(-1), _last_clicked(-1), _needs_relayout(true), _collapsed(false), _was_collapsed(false)
{
  _background_pattern= 0;
  _line_pattern= 0;
  _selection_image = NULL;
  _timeout = 0;
  switch (type)
  {
    case VerticalIconSwitcher:
#ifdef _WIN32
      if (base::Color::get_active_scheme() != base::ColorSchemeStandardWin7)
      {
        _colors[TabInactiveBackground]= base::Color::get_application_color(base::AppColorPanelHeader, false);
        _colors[TabInactiveForeground]= base::Color::get_application_color(base::AppColorPanelHeader, true);
        _colors[TabActiveBackground]= base::Color::get_application_color(base::AppColorPanelHeaderFocused, false);
        _colors[TabActiveForeground]= base::Color::get_application_color(base::AppColorPanelHeaderFocused, true);
      }
      else
      {
        _colors[TabInactiveBackground]= base::Color::get_application_color(base::AppColorPanelHeader, false);
        _colors[TabInactiveForeground]= base::Color::get_application_color(base::AppColorPanelHeader, true);
        _colors[TabActiveBackground]= base::Color::get_application_color(base::AppColorMainBackground, false);
        _colors[TabActiveForeground]= base::Color::get_application_color(base::AppColorPanelHeader, true);
      }
#else
      _colors[TabInactiveBackground]= Color(0x48/255.0, 0x48/255.0, 0x48/255.0, 1);
      _colors[TabActiveBackground]= Color(0x26/255.0, 0x26/255.0, 0x26/255.0, 1);
      _colors[TabActiveForeground]= Color(1, 1, 1, 1);
      _colors[TabInactiveForeground]= Color(0.6, 0.6, 0.6, 1);
#endif
      _selection_image = cairo_image_surface_create_from_png(mforms::App::get()->get_resource_path("output_type-item_selected.png").c_str());

      set_size(VERTICAL_STYLE_WIDTH, -1);
      break;
  }
}

//--------------------------------------------------------------------------------------------------

TabSwitcher::~TabSwitcher()
{
  if (_timeout)
    mforms::Utilities::cancel_timeout(_timeout);
  
  for (std::vector<TabItem*>::iterator iter= _items.begin(); iter != _items.end(); ++iter)
    delete *iter;
  
  if (_selection_image)
    cairo_surface_destroy(_selection_image);
  
  destroy_patterns();
}

//--------------------------------------------------------------------------------------------------

/**
 * Prepare patterns we need for drawing.
 */
void TabSwitcher::prepare_patterns()
{
  if (_background_pattern == NULL)
  {
    _background_pattern = cairo_pattern_create_linear(0, 0, 0, get_height());
    cairo_pattern_add_color_stop_rgba(_background_pattern, 0, 1, 1, 1, 0);
    cairo_pattern_add_color_stop_rgba(_background_pattern, 1, 1, 1, 1, 1);
  }
  if (_line_pattern == NULL)
  {
    Color line_color= _colors[TabLineColor];
    
    _line_pattern = cairo_pattern_create_linear(0, 0, 0, get_height());
    cairo_pattern_add_color_stop_rgba(_line_pattern, 0, line_color.red, line_color.green,
                                      line_color.blue, 0);
    cairo_pattern_add_color_stop_rgba(_line_pattern, 1, line_color.red, line_color.green,
                                      line_color.blue, line_color.alpha);
  }
}

//--------------------------------------------------------------------------------------------------

/**
 * If any pattern is allocated, destroy it.
 */
void TabSwitcher::destroy_patterns()
{
  if (_background_pattern != NULL)
    cairo_pattern_destroy(_background_pattern);
  if (_line_pattern != NULL)
    cairo_pattern_destroy(_line_pattern);
}

//--------------------------------------------------------------------------------------------------

int TabSwitcher::get_preferred_height() 
{
  return INITIAL_TAB_HEIGHT; 
}

//--------------------------------------------------------------------------------------------------

void TabSwitcher::attach_to_tabview(TabView *tabView)
{
  _tabView= tabView;
  set_needs_relayout();
  
  scoped_connect(_tabView->signal_tab_changed(),boost::bind(&TabSwitcher::tab_changed, this));
}

//--------------------------------------------------------------------------------------------------

void TabSwitcher::set_collapsed(bool flag)
{
  if (_collapsed != flag)
  {
    _collapsed = flag;
    if (_type == VerticalIconSwitcher)
    {
      set_size(_collapsed ? 5 : VERTICAL_STYLE_WIDTH, -1);
      _signal_collapse_changed();
    }
    set_layout_dirty();
  }
}

//--------------------------------------------------------------------------------------------------

int TabSwitcher::add_item(const std::string &title, const std::string &sub_title,
                          const std::string &icon_path, const std::string &alt_icon_path)
{
  TabItem *item = new TabItem();

  item->title= title;
  item->sub_title= sub_title;
  item->icon= cairo_image_surface_create_from_png(icon_path.c_str());
  if (item->icon && cairo_surface_status(item->icon) != CAIRO_STATUS_SUCCESS)
  {
    cairo_surface_destroy(item->icon);
    item->icon= NULL;
  }

  item->alt_icon= cairo_image_surface_create_from_png(alt_icon_path.c_str());
  if (item->alt_icon && cairo_surface_status(item->alt_icon) != CAIRO_STATUS_SUCCESS)
  {
    cairo_surface_destroy(item->alt_icon);
    item->alt_icon= NULL;
  }

  _items.push_back(item);

  if (_selected == -1)
    set_selected(_items.size() - 1);
  
  set_needs_relayout();
  return _items.size() - 1;
}

//--------------------------------------------------------------------------------------------------

/**
 * Replaces the icon pair for the given item with the new values.
 */
void TabSwitcher::set_icon(int index, const std::string &icon_path, const std::string &alt_icon_path)
{
  if (index >= 0 && index < (int) _items.size())
  {
    TabItem *item = _items[index];

    if (item->icon != NULL)
      cairo_surface_destroy(item->icon);
    item->icon= cairo_image_surface_create_from_png(icon_path.c_str());
    if (item->icon && cairo_surface_status(item->icon) != CAIRO_STATUS_SUCCESS)
    {
      cairo_surface_destroy(item->icon);
      item->icon= NULL;
    }

    if (item->alt_icon != NULL)
      cairo_surface_destroy(item->alt_icon);
    item->alt_icon= cairo_image_surface_create_from_png(alt_icon_path.c_str());
    if (item->alt_icon && cairo_surface_status(item->alt_icon) != CAIRO_STATUS_SUCCESS)
    {
      cairo_surface_destroy(item->alt_icon);
      item->alt_icon= NULL;
    }

    set_needs_relayout();
  }
}

//--------------------------------------------------------------------------------------------------

void TabSwitcher::set_selected(int index)
{
  _selected= index;
  if (_tabView != NULL)
    _tabView->set_active_tab(index);
  set_needs_repaint();
}

//--------------------------------------------------------------------------------------------------

void TabSwitcher::set_needs_relayout()
{
  _needs_relayout= true;
  set_needs_repaint();
}

//--------------------------------------------------------------------------------------------------

void TabSwitcher::set_layout_dirty()
{
  DrawBox::set_layout_dirty(true);
  set_needs_relayout();
}

//--------------------------------------------------------------------------------------------------

void TabSwitcher::layout(cairo_t *cr)
{
  if (_needs_relayout)
  {
  }
}

//--------------------------------------------------------------------------------------------------

bool TabSwitcher::mouse_down(mforms::MouseButton button, int x, int y)
{
  if (!DrawBox::mouse_down(button, x, y))
  {
    // For now ignore which button was pressed.
    _last_clicked = index_from_point(x, y);
  }

  return true;
}

//--------------------------------------------------------------------------------------------------

bool TabSwitcher::mouse_up(mforms::MouseButton button, int x, int y)
{
  return DrawBox::mouse_up(button, x, y);
}

//--------------------------------------------------------------------------------------------------

bool TabSwitcher::mouse_click(mforms::MouseButton button, int x, int y)
{
  bool handled = DrawBox::mouse_click(button, x, y);

  // Don't change anything if the user clicked outside of any tab.
  if (!handled && _last_clicked > -1 && _last_clicked == index_from_point(x, y))
  {
    set_selected(_last_clicked);
    _signal_changed();
    handled = true;
  }
  return handled;
}

//--------------------------------------------------------------------------------------------------

bool TabSwitcher::mouse_enter()
{
  _was_collapsed = _collapsed;
  if (_collapsed)
    set_collapsed(false);
  return true;
}

//--------------------------------------------------------------------------------------------------

bool TabSwitcher::collapse()
{
  _timeout = 0;
  set_collapsed(true);
  return false;
}

//--------------------------------------------------------------------------------------------------

bool TabSwitcher::mouse_leave()
{
  if (_was_collapsed)
  {
    _was_collapsed = false;
    _timeout = mforms::Utilities::add_timeout(0.3f, boost::bind(&TabSwitcher::collapse, this));
  }
  return true;
}

//--------------------------------------------------------------------------------------------------

void TabSwitcher::repaint(cairo_t *cr, int areax, int areay, int areaw, int areah)
{
  repaint_vertical(cr, areax, areay, areaw, areah);
}

//--------------------------------------------------------------------------------------------------

void TabSwitcher::repaint_vertical(cairo_t *cr, int areax, int areay, int areaw, int areah)
{
  Color color;
  
  layout(cr);
  prepare_patterns();
  
  cairo_save(cr);

  color = _colors[TabInactiveBackground];
  cairo_set_source_rgba(cr, color.red, color.green, color.blue, 1);
  cairo_paint(cr);

  float backing_scale = mforms::App::get()->backing_scale_factor();
  const int font_size = 10;
  
  cairo_select_font_face(cr, TAB_FONT, CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size(cr, font_size);

  double y = 0;
  double iy = 0;
  int ii = 0;
  for (std::vector<TabItem*>::iterator i = _items.begin(); i != _items.end(); ++i, ++ii)
  {
    double ih = cairo_image_surface_get_height((*i)->icon) / backing_scale;
    
    if (_selected == ii)
    {
      color = _colors[TabActiveBackground];
      cairo_set_source_rgba(cr, color.red, color.green, color.blue, 1);
      cairo_rectangle(cr, 0, y, VERTICAL_STYLE_WIDTH, VERTICAL_STYLE_HEIGHT);
      cairo_fill(cr);

      cairo_save(cr);
      cairo_scale(cr, 1/backing_scale, 1/backing_scale);
      cairo_set_source_surface(cr, _selection_image,
                               0, iy + (VERTICAL_STYLE_WIDTH - ih)/2 + cairo_image_surface_get_height(_selection_image)/2);
      cairo_paint(cr);
      cairo_restore(cr);
    }

    cairo_save(cr);
    cairo_scale(cr, 1/backing_scale, 1/backing_scale);
    cairo_set_source_surface(cr, (*i)->icon,
                             (backing_scale*VERTICAL_STYLE_WIDTH - (VERTICAL_STYLE_WIDTH/64.0)*cairo_image_surface_get_width((*i)->icon))/2,
                             iy + (VERTICAL_STYLE_WIDTH - ih)/2 - (backing_scale > 1 ? 0 : font_size));
    if (_selected == ii)
    {
      cairo_paint(cr);
      color = _colors[TabActiveForeground];
    }
    else
    {
      cairo_paint_with_alpha(cr, 0.4);
      color = _colors[TabInactiveForeground];
    }
    cairo_restore(cr);

    cairo_set_source_rgba(cr, color.red, color.green, color.blue, 1);

    std::string::size_type p = (*i)->title.find('\n');
    if (p == std::string::npos)
    {
      cairo_text_extents_t ext;
      cairo_text_extents(cr, (*i)->title.c_str(), &ext);

      cairo_move_to(cr, (VERTICAL_STYLE_WIDTH - ext.width)/2, y + ih);
      cairo_show_text(cr, (*i)->title.c_str());
    }
    else
    {
      std::string l1 = (*i)->title.substr(0, p);
      std::string l2 = (*i)->title.substr(p+1);
      cairo_text_extents_t ext1, ext2;
      cairo_text_extents(cr, l1.c_str(), &ext1);
      cairo_text_extents(cr, l2.c_str(), &ext2);
      
      cairo_move_to(cr, (VERTICAL_STYLE_WIDTH - ext1.width)/2, y + ih + 4 - (font_size + ext1.y_bearing) + (VERTICAL_STYLE_WIDTH - ih)/2);
      cairo_show_text(cr, l1.c_str());
      cairo_stroke(cr);
      cairo_move_to(cr, (VERTICAL_STYLE_WIDTH - ext2.width)/2, y + ih + 4 + font_size - (font_size + ext2.y_bearing) + (VERTICAL_STYLE_WIDTH - ih)/2);
      cairo_show_text(cr, l2.c_str());
      cairo_stroke(cr);
    }
    y += VERTICAL_STYLE_HEIGHT;
    iy += VERTICAL_STYLE_HEIGHT * backing_scale;
  }
  
  cairo_restore(cr);
}

//--------------------------------------------------------------------------------------------------

/**
 * Returns the index of the tab item at the given position or -1 if there is none.
 *
 * The coordinates must be given in local space.
 */
int TabSwitcher::index_from_point(int x, int y)
{
  if (_items.size() == 0 || x < 0 || x > get_width() || y < 0 || y > get_height())
    return -1;
  
  switch (_type)
  {
    case VerticalIconSwitcher:
    {
      for (size_t i = 0; i < _items.size(); i++)
      {
        if (y < (int)(i+1) * VERTICAL_STYLE_HEIGHT)
          return i;
      }
      break;
    }
    /*
  float offset= 0.5 + SIDE_PADDING;
  if (x < offset)
    return -1;
  
  int i= 0;
  for (std::vector<TabItem*>::iterator iter = _items.begin(); iter != _items.end(); ++iter, ++i)
  {
    if (offset <= x && x <= offset + (*iter)->width)
      return i;
    offset += (*iter)->width;
  }*/
  }
  return -1;
}

//--------------------------------------------------------------------------------------------------

void TabSwitcher::tab_changed()
{
  _selected = _tabView->get_active_tab();
  set_needs_repaint();
}
